// Copyright (c) 2002 Andreas Jnsson of www.AngelCode.com
#include "datatypes.h"
#include <math.h>
#include <assert.h>
#include "analyze.h"

void WaveletAnalyze(int N, float *xin, float even, float odd, float *yout)
{
	float *xtmp1 = new float[N];
	float *xtmp2 = new float[N];
	float *xswap;

	// Copy input to temporary buffer
	for( int i = 0; i < N; i++ )
		xtmp1[i] = xin[i];

	int n = N;
	while( n > 1 )
	{
		// Analyze signal
		WaveletAnalyzeStep(n, xtmp1, even, odd, xtmp2);

		// Copy wavelet coefficients to output
		for( int i = n/2; i < n; i++ )
			yout[i] = xtmp2[i];

		xswap = xtmp2; xtmp2 = xtmp1; xtmp1 = xswap;

		n /= 2;
	}

	// Copy the last low frequency coefficient to output
	yout[0] = xtmp1[0];

	delete[] xtmp1;
	delete[] xtmp2;
}

void WaveletSynthesize(int N, float *yin, float even, float odd, float *xout)
{
	float *xtmp1 = new float[N];
	float *xtmp2 = new float[N];
	float *xswap;

	// Copy input to temporary buffer
	for( int i = 0; i < 2; i++ )
		xtmp1[i] = yin[i];

	int n = 2;
	while( n < N )
	{
		// Analyze signal
		WaveletSynthesizeStep(n, xtmp1, even, odd, xtmp2);

		xswap = xtmp2; xtmp2 = xtmp1; xtmp1 = xswap;

		// Copy wavelet coefficients to temporary buffer
		for( int i = n; i < n*2; i++ )
			xtmp1[i] = yin[i];

		n *= 2;
	}

	// Analyze the last level to the output
	WaveletSynthesizeStep(N, xtmp1, even, odd, xout);

	delete[] xtmp1;
	delete[] xtmp2;
}

void WaveletAnalyzeStep(int N, float *xin, float even, float odd, float *yout)
{
	// yhigh[k] = sum_n(x[n]*g[2k-n])
	// ylow[k]  = sum_n(x[n]*h[2k-n])

	// This example matrix multiplication shows what is going on

	// [yl0] = [h3 h2 h1 h0            ][x0]
	// [yl1] = [      h3 h2 h1 h0      ][x1]
	// [yl2] = [            h3 h2 h1 h0][x2]
	// [yl3] = [h1 h0             h3 h2][x3]
	// [yh0] = [g3 g2 g1 g0            ][x4]
	// [yh1] = [      g3 g2 g1 g0      ][x5]
	// [yh2] = [            g3 g2 g1 g0][x6]
	// [yh3] = [g1 g0             g3 g2][x7]

	// Initialize the scaling function and wavelet function and 
	// normalize so that the integral of the scaling function is 2/sqrt(2).
	// This normalization is done so that the energy of the wavelet 
	// function is equal on each step. 
	const float norm = 1.0f/1.4142135623730951f;
	float h[4] = {even*norm, (1-odd)*norm, (1-even)*norm, odd*norm};
	float g[4] = {h[3], -h[2], h[1], -h[0]};

	float *yhigh = &yout[N/2];
	float *ylow  = yout;

	if( N > 2 )
	{
		for( int i = 0, o = 0; i < N-3; i += 2, o++ )
		{
			yhigh[o] = xin[i+0]*g[3] + xin[i+1]*g[2] + xin[i+2]*g[1] + xin[i+3]*g[0];
			ylow[o]  = xin[i+0]*h[3] + xin[i+1]*h[2] + xin[i+2]*h[1] + xin[i+3]*h[0];
		}

		// The last one uses wrap around to not loose information
		yhigh[o] = xin[i+0]*g[3] + xin[i+1]*g[2] + xin[0]*g[1] + xin[1]*g[0];
		ylow[o]  = xin[i+0]*h[3] + xin[i+1]*h[2] + xin[0]*h[1] + xin[1]*h[0];
	}
	else
	{
		// Use wrap around to not loose information
		*yhigh = xin[0]*g[3] + xin[1]*g[2] + xin[0]*g[1] + xin[1]*g[0];
		*ylow  = xin[0]*h[3] + xin[1]*h[2] + xin[0]*h[1] + xin[1]*h[0];
	}
}

void WaveletSynthesizeStep(int N, float *yin, float even, float odd, float *xout)
{
	// x[n] = sum_k(yhigh[k]*g[2k-n] + ylow[k]*h[2k-n])

	// This example matrix multiplication shows what is going on.
	// Note that this matrix is the transposed version of the matrix above.

	// [x0] = [h3       h1 g3       g1][yl0]
	// [x1] = [h2       h0 g2       g0][yl1]
	// [x2] = [h1 h3       g1 g3      ][yl2]
	// [x3] = [h0 h2       g0 g2      ][yl3]
	// [x4] = [   h1 h3       g1 g3   ][yh0]
	// [x5] = [   h0 h2       g0 g2   ][yh1]
	// [x6] = [      h1 h3       g1 g3][yh2]
	// [x7] = [      h0 h2       g0 g2][yh3]

	// Initialize the synthesize filters. Note that these are 
	// the same as the analyzis filters.
	const float norm = 1.0f/1.4142135623730951f;
	float h[4] = {even*norm, (1-odd)*norm, (1-even)*norm, odd*norm};
	float g[4] = {h[3], -h[2], h[1], -h[0]};

	float *yhigh = &yin[N/2];
	float *ylow  = yin;

	if( N > 2 )
	{
		int i = N/2-1, o;
	
		// Use wrap around to not loose information
		xout[0] = ylow[i]*h[1] + ylow[0]*h[3] + yhigh[i]*g[1] + yhigh[0]*g[3]; 
		xout[1] = ylow[i]*h[0] + ylow[0]*h[2] + yhigh[i]*g[0] + yhigh[0]*g[2]; 

		for( o = 2, i = 0; o < N; o += 2, i++ )
		{
			xout[o  ] = ylow[i]*h[1] + ylow[i+1]*h[3] + yhigh[i]*g[1] + yhigh[i+1]*g[3]; 
			xout[o+1] = ylow[i]*h[0] + ylow[i+1]*h[2] + yhigh[i]*g[0] + yhigh[i+1]*g[2]; 
		}
	}
	else
	{
		// Use wrap around to not loose information
		xout[0] = ylow[0]*h[1] + ylow[0]*h[3] + yhigh[0]*g[1] + yhigh[0]*g[3]; 
		xout[1] = ylow[0]*h[0] + ylow[0]*h[2] + yhigh[0]*g[0] + yhigh[0]*g[2]; 
	}
}

// This function generates an array that show how the  
// scaling function looks as a continuos function 
void GenerateScalingFunction(float even, float odd, int length, float *out)
{
	assert(out != 0);

	// Working buffers
	float *t1 = new float[length];
	float *t2 = new float[length];
	float *ts;
	int l1 = 5;
	int l2;

	float evenc = 1 - even;
	float oddc  = 1 - odd;

	// Starting coefficients
	t1[0] = t1[1] = t1[3] = t1[4] = 0;
	t1[2] = 1;

	while( l1 < length/2 )
	{
		l2 = 2*l1 - 2;
		int i1 = 0, i2 = 0;
		for( ; i1 < l1 - 1; i1++ )
		{
			t2[i2++] = evenc * t1[i1] + even * t1[i1+1];
			t2[i2++] =   odd * t1[i1] + oddc * t1[i1+1];
		}

		// Swap the work buffers
		ts = t2; t2 = t1; t1 = ts;
		l1 = l2;
	}

	// Copy to the output
	for( int n = 0; n < (length - l1)/2; n++ )
		out[n] = 0;

	for( int i = 0; i < l1; i++, n++ )
		out[n] = t1[i];

	for( ; n < length; n++ )
		out[n] = 0;

	// Clean up
	delete[] t1;
	delete[] t2;
}

// This function generates an array that show how the 
// wavelet function looks as a continuos function
void GenerateWaveletFunction(float even, float odd, int length, float *out)
{
	assert(out != 0);

	// Working buffers
	float *t1 = new float[length];
	float *t2 = new float[length];
	float *ts;
	int l1 = 8;
	int l2;

	float evenc = 1 - even;
	float oddc  = 1 - odd;

	// Starting coefficients
	t1[0] = t1[1] = t1[6] = t1[7] = 0;
	t1[2] = odd;
	t1[3] = -evenc;
	t1[4] = oddc;
	t1[5] = -even;

	while( l1 < length/2 )
	{
		l2 = 2*l1 - 2;
		int i1 = 0, i2 = 0;
		for( ; i1 < l1 - 1; i1++ )
		{
			t2[i2++] = evenc * t1[i1] + even * t1[i1+1];
			t2[i2++] =   odd * t1[i1] + oddc * t1[i1+1];
		}

		// Swap the work buffers
		ts = t2; t2 = t1; t1 = ts;
		l1 = l2;
	}

	// Copy to the output
	for( int n = 0; n < (length - l1)/2; n++ )
		out[n] = 0;

	for( int i = 0; i < l1; i++, n++ )
		out[n] = t1[i];

	for( ; n < length; n++ )
		out[n] = 0;

	// Clean up
	delete[] t1;
	delete[] t2;
}