// Copyright (c) 2002 Andreas Jnsson of www.AngelCode.com
#include "mywindow.h"
#include "resource.h"
#include <math.h>
#include "analyze.h"
#include "stdlib.h"

cMyWindow::cMyWindow() : cSimpleWindow()
{
}

int cMyWindow::Create(int x, int y)
{
	int cyMenu = GetSystemMetrics(SM_CYMENU);
	
	// Create the simple window
	int r = cSimpleWindow::Create(x, y + cyMenu, "Discreet Wavelet Transform", 0);
	if( r < 0 ) return r;

	// Set the menu
	HMENU hMenu = LoadMenu(::GetModuleHandle(0), MAKEINTRESOURCE(IDR_MENU));
	::SetMenu(hWnd, hMenu);

	// Note that the window will automatically destroy the 
	// menu when destroyed so we don't need to keep track of 
	// this resource

	// Set some initial settings
	showBackground = false;
	even = odd = 0;
	::SendMessage(hWnd, WM_COMMAND, ID_OPTIONS_USESAMPLE_1, 0);
	::SendMessage(hWnd, WM_COMMAND, ID_VIEW_ORIGINAL, 0);

	return 0;
}

//=============================================================================

void cMyWindow::ComputeSample(int wave)
{
	if( wave == 0 )
	{
		float a = 0;
		float d = 0.01f;
		for( int n = 0; n < DATA_LENGTH; n++ )
		{
			original[n] = sinf(a);
			a += d;
			d += 0.005f;
		}
	}
	else if( wave == 1 )
	{
		float a = 0;
		for( int n = 0; n < DATA_LENGTH/4; n++ )
		{
			original[n] = sinf(a);
			a += 0.4f;
		}
		for( ; n < DATA_LENGTH/2; n++ )
		{
			original[n] = sinf(a);
			a += 0.5f;
		}
		for( ; n < 3*DATA_LENGTH/4; n++ )
		{
			original[n] = sinf(a);
			a += 0.3333f;
		}
		for( ; n < DATA_LENGTH; n++ )
		{
			original[n] = sinf(a);
			a += 0.25f;
		}
	}
	else if( wave == 2 )
	{
		float a1 = 0, a2 = 0;
		for( int n = 0; n < DATA_LENGTH; n++ )
		{
			original[n] = sinf(a1)*(sinf(a2)/4 + 0.5f);
			a1 += 0.5f;
			a2 += 0.05f;
		}
	}
	else if( wave == 3 )
	{
		for( int n = 0; n < DATA_LENGTH; n++ )
		{
			original[n] = (float)(rand())/RAND_MAX*2-1;
		}
	}
}

void cMyWindow::ShowGraph(int graph)
{
	showBackground = false;
	if( graph == 0 )
	{
		for( int n = 0; n < DATA_LENGTH; n++ )
			data[n] = original[n];
	}
	else if( graph == 1 )
	{
		// Compute the wavelet transform of the sample
		WaveletAnalyze(DATA_LENGTH, original, even, odd, data);
		showBackground = true;
	}
	else if( graph == 2 )
	{
		// Compute the wavelet transform of the sample and then return
		float dwt[DATA_LENGTH];
		WaveletAnalyze(DATA_LENGTH, original, even, odd, dwt);
		WaveletSynthesize(DATA_LENGTH, dwt, even, odd, data);
	}
	else if( graph == 3 )
	{
		// Take the different of graph 2 and the original data
		float synthesized[DATA_LENGTH];
		WaveletAnalyze(DATA_LENGTH, original, even, odd, data);
		WaveletSynthesize(DATA_LENGTH, data, even, odd, synthesized);
		for( int n = 0; n < DATA_LENGTH; n++ )
			data[n] = synthesized[n] - original[n];
	}
	else if( graph == 4 )
	{
		GenerateWaveletFunction(even, odd, DATA_LENGTH, data);

		// Scale the function so that all is visible
		for( int n = 0; n < DATA_LENGTH; n++ )
			data[n] *= 0.5f;
	}
	else if( graph == 5 )
	{
		GenerateScalingFunction(even, odd, DATA_LENGTH, data);

		// Scale the function so that all is visible
		for( int n = 0; n < DATA_LENGTH; n++ )
			data[n] *= 0.5f;
	}

	::InvalidateRect(hWnd, 0, TRUE);
}

void cMyWindow::SetWavelet(int wl)
{
	if( wl == 0 )
	{
		// Haar wavelet : Orthogonal => Perfect reconstruction
		even = odd = 0;
	}
	else if( wl == 1 )
	{
		// Daubechies : Orthogonal => Perfect reconstruction
		even  = float((1+sqrt(3))/4);
		odd   = 0.5f - even;
	}
	else if( wl == 2 )
	{
		// Spline : Not orthogonal => Imperfect reconstruction
		even = 0.25f;
		odd  = 0.25f;
	}
	else if( wl == 3 )
	{
		// Linear : Not orthogonal => Imperfect reconstruction
		even = 0.5f;
		odd  = 0;
	}
}

//=============================================================================

void cMyWindow::OnPaint()
{
	HDC dc;
	PAINTSTRUCT ps; 

	dc = ::BeginPaint(hWnd, &ps);

	DrawGraph(dc);

	::EndPaint(hWnd, &ps);
}

void cMyWindow::DrawGraph(HDC dc)
{
	RECT rect;

	GetClientRect(&rect);
	float width = (float)rect.right;
	float cy = (float)rect.bottom/2; 

	// Draw background in gray to highlight the different levels of DWT coefficients
	if( showBackground )
	{
		HBRUSH gray = ::CreateSolidBrush(RGB(80,80,80));

		int x1 = rect.right/2;
		int x2 = rect.right;
		int n = DATA_LENGTH;
		while( n > 0 )
		{
			RECT rc = {x1, 0, x2, rect.bottom};
			::FillRect(dc, &rc, gray);

			n /= 4;
			x2 = x1/2;
			x1 /= 4;
		}

		::DeleteObject(gray);
	}

	// Draw the middle red line
	HPEN redPen, pen, oldPen;

	redPen = ::CreatePen(PS_SOLID, 1, RGB(255,0,0));
	oldPen = (HPEN)::SelectObject(dc, redPen);

	::MoveToEx(dc, 0, (int)cy, NULL);
	::LineTo(dc, (int)width-1, (int)cy);

	// Draw the graph
	pen = ::CreatePen(PS_SOLID, 1, RGB(255,255,255));
	::SelectObject(dc, pen);

	float samplesPerPixel = (DATA_LENGTH-1)/width;
	float prev = 0;
	for( int x = 0; x < width; x++ )
	{
		int n = (int)(x*samplesPerPixel);
		float min = data[n], max = data[n];
		for( ; n < (int)((x+1)*samplesPerPixel); n++ )
		{
			if( data[n] < min ) min = data[n];
			if( data[n] > max ) max = data[n];
		}

		::MoveToEx(dc, x-1, (int)(cy - prev*cy), NULL);
		::LineTo(dc, x, (int)(cy - (min+max)/2*cy));
		prev = (min+max)/2;

		::MoveToEx(dc, x, (int)(cy - min*cy), NULL);
		::LineTo(dc, x, (int)(cy - max*cy));
	}
	
	::SelectObject(dc, oldPen);
	::DeleteObject(pen);
}

//=============================================================================

LRESULT cMyWindow::MsgProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	switch(msg)
	{
	case WM_COMMAND:
		switch(wParam)
		{
		case IDC_EXIT:
			::DestroyWindow(hWnd);
			return 0;

		case ID_VIEW_ORIGINAL:
		case ID_VIEW_WAVELETTRANSFORM:
		case ID_VIEW_SYNTHESIZED:
		case ID_VIEW_SAMPLEERRORS:
		case ID_VIEW_WAVELET:
		case ID_VIEW_SCALINGFUNCTION:
			ShowGraph(wParam - ID_VIEW_ORIGINAL);
			return 0;

		case ID_OPTIONS_USESAMPLE_1:
		case ID_OPTIONS_USESAMPLE_2:
		case ID_OPTIONS_USESAMPLE_3:
		case ID_OPTIONS_USESAMPLE_4:
			ComputeSample(wParam - ID_OPTIONS_USESAMPLE_1);
			ShowGraph(0);
			return 0;

		case ID_OPTIONS_USEWAVELET_HAAR:
		case ID_OPTIONS_USEWAVELET_DAUBECHIES:
		case ID_OPTIONS_USEWAVELET_SPLINE:
		case ID_OPTIONS_USEWAVELET_LINEAR:
			SetWavelet(wParam - ID_OPTIONS_USEWAVELET_HAAR);
			ShowGraph(1);
			return 0;

		case ID_HELP_ABOUT:
			::DialogBox(::GetModuleHandle(0),MAKEINTRESOURCE(IDD_ABOUT),hWnd,AboutProc);
			return 0;
		}
		break;

	case WM_PAINT:
		OnPaint();
		return 0;
	}

	return cSimpleWindow::MsgProc(hWnd, msg, wParam, lParam);
}

BOOL CALLBACK cMyWindow::AboutProc(HWND hwndDlg, UINT message, WPARAM wParam, LPARAM lParam) 
{ 
	// Don't do anything
	switch( message )
	{
	case WM_CLOSE:
		EndDialog(hwndDlg, 0);
		return TRUE;

	case WM_COMMAND:
		if( wParam == IDOK )
		{
			EndDialog(hwndDlg, 0);
			return TRUE;
		}
		break;
	}

	return FALSE;
} 