//-----------------------------------------------------------------------------
// Copyright (c) 2002 Andreas Jnsson
//-----------------------------------------------------------------------------
#include "simplewindow.h"
#include "errorcodes.h"

//=============================================================================

cSimpleWindow::cSimpleWindow()
{
	hWnd        = 0;
	windowClass = 0;
}

//=============================================================================

int cSimpleWindow::Create(int width, int height, const char *title, DWORD flags)
{
	int r;
	r = RegClass(0, 0, 0, 0, 0, 0);
	if( FAILED(r) )
		return r;

	r = CreateWnd(width, height, title, flags);
	if( FAILED(r) )
		return r;

	return SC_OK;
}

//=============================================================================

int cSimpleWindow::RegClass(const char *newMenuName, const char *accel,
						   HBRUSH newBgBrush, HICON newIcon, 
						   HCURSOR newCursor, const char *newClassName)
{
	// Remember the class name and menu name
	char *className;
	if( newClassName != 0 )
		className = (char *)newClassName;
	else
		className = "AngelCode simple window class";

	// Get the module handle
	HINSTANCE hInst = ::GetModuleHandle(0);

	// See if the class has been registered before
	WNDCLASSEX wndClass;
	if( GetClassInfoEx(hInst, className, &wndClass) )
		return SC_OK;

	// Register the new class
	isUsingMenu = (newMenuName != 0);

	HBRUSH  hBgBrush   = newBgBrush;
	HICON   hIcon      = newIcon;
	HCURSOR hCursor    = newCursor;
	HICON   hIconSmall = 0;

	if( hIcon == 0 )    hIcon    = ::LoadIcon(NULL, IDI_WINLOGO);
	if( hCursor == 0 )  hCursor  = ::LoadCursor(NULL, IDC_ARROW);
	if( hBgBrush == 0 ) hBgBrush = (HBRUSH)::GetStockObject(BLACK_BRUSH);

	// Register the window class
	wndClass.cbSize        = sizeof(WNDCLASSEX);
	wndClass.style         = CS_HREDRAW | CS_VREDRAW | CS_OWNDC;
	wndClass.lpfnWndProc   = (WNDPROC)WndProc;
	wndClass.cbClsExtra    = 0;
	wndClass.cbWndExtra    = 0;
	wndClass.hInstance     = hInst;
	wndClass.hIcon         = hIcon;
	wndClass.hCursor       = hCursor;
	wndClass.hbrBackground = hBgBrush;
	wndClass.lpszMenuName  = newMenuName;
	wndClass.lpszClassName = className;
	wndClass.hIconSm       = hIconSmall;

	windowClass = ::RegisterClassEx( &wndClass );
	if( !windowClass )
		return EC_FailedToRegisterWindowClass;

	// Load the accelerator table
	if( accel != 0 )
	{
		hAccel = ::LoadAccelerators(hInst, accel);
		if( hAccel == 0 )
			return EC_FailedToLoadAccelerators;
	}
	else
		hAccel = 0;

	return SC_OK;
}

//=============================================================================

int cSimpleWindow::CreateWnd(int _width, int _height, const char *title, DWORD flags)
{
	if( windowClass == 0 )
		return EC_NoWindowClass;

	HINSTANCE hInst = ::GetModuleHandle(0);
	isFullscreen    = flags & WCF_FULLSCREEN ? true : false;
 
    // Setup our window parameters
	int posX, posY, width, height;
	DWORD exStyle, style;
	if( isFullscreen )
	{
		exStyle = WS_EX_TOPMOST; // This lets our window cover the whole screen
		style   = WS_POPUP;      // We don't need caption and system 
		                         // menu when in fullscreen mode

		// Set the window to cover the whole screen
		posX   = 0;
		posY   = 0;
		width  = ::GetSystemMetrics(SM_CXSCREEN);
		height = ::GetSystemMetrics(SM_CYSCREEN);
	}
	else
	{
		exStyle = 0;
		if( flags & WCF_NORESIZE )
			style = WS_OVERLAPPED|WS_CAPTION|WS_SYSMENU|WS_DLGFRAME; 
		else
			style = WS_OVERLAPPEDWINDOW; // This is a normal window

		// Make sure the client area is the size we want
		RECT rect = {0, 0, _width, _height};
		::AdjustWindowRect(&rect, style, isUsingMenu);

		posX   = CW_USEDEFAULT; // Let Windows position our window
		posY   = CW_USEDEFAULT;
		width  = rect.right - rect.left;
		height = rect.bottom - rect.top;
	}


    // Create the render window
	// Set lpParam to point to our application so that the window procedure
	// knows which class to call when handling window messages
    hWnd = ::CreateWindowEx( exStyle, (char *)windowClass, 
		                     title, style, posX, posY,
						     width,height,NULL,NULL,hInst,this ); 
	if( hWnd == NULL )
		return EC_FailedToCreateWindow;

	return SC_OK;
}

//=============================================================================

int cSimpleWindow::Show()
{
	// Show the window
	::ShowWindow( hWnd, SW_SHOW );
    ::UpdateWindow( hWnd );

	return SC_OK;
}

//=============================================================================

// Static msg handler which passes messages to the application class.
LRESULT CALLBACK cSimpleWindow::WndProc( HWND hWnd, UINT msg, 
								   WPARAM wParam, LPARAM lParam )
{
	cSimpleWindow *wnd;

	if( msg == WM_NCCREATE )
	{
		// Extract the window pointer parameter.
		CREATESTRUCT *pCS = (CREATESTRUCT *)lParam;
		wnd = (cSimpleWindow *)pCS->lpCreateParams;

		if( wnd )
		{
			// Save it in the window handle.
			::SetWindowLong(hWnd, GWL_USERDATA, (LONG)wnd);
		}
	}
	else
	{
		// Extract the object from the window handle.
		wnd = (cSimpleWindow *)::GetWindowLong(hWnd, GWL_USERDATA);
	}

	if( wnd )
	{
		// Let the object handle it.
		return wnd->MsgProc(hWnd, msg, wParam, lParam);
	}
	else
	{
		// Just do the default if the window is not created yet.
		return ::DefWindowProc(hWnd, msg, wParam, lParam);
	}
}


//=============================================================================

// If wait is true then the function will wait until a message arrives.
// The function returns true if it receives WM_QUIT.
bool cSimpleWindow::CheckMessages(bool wait)
{
    MSG msg;
	cSimpleWindow *wnd;
    
	if( wait )
		::WaitMessage();

	while( ::PeekMessage(&msg, NULL, 0, 0, PM_REMOVE) )
    {
		if( msg.message == WM_QUIT )
			return true;

		wnd = (cSimpleWindow *)::GetWindowLong(msg.hwnd, GWL_USERDATA);
		if( wnd )
			wnd->DispatchMessage(&msg);
		else
			::DispatchMessage(&msg);
	}

    return false;
}

//=============================================================================

void cSimpleWindow::DispatchMessage(MSG *msg)
{
	if( hAccel && ::TranslateAccelerator(hWnd, hAccel, msg) )
		return;

    ::TranslateMessage(msg);
	::DispatchMessage(msg);
}

//=============================================================================

HWND cSimpleWindow::GetHandle()
{
	return hWnd;
}

//=============================================================================

void cSimpleWindow::GetClientRect(RECT *rect)
{
	::GetClientRect(hWnd, rect);
}

//=============================================================================

LRESULT cSimpleWindow::MsgProc( HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam )
{
	if( msg == WM_DESTROY )
	{
		::PostQuitMessage(0);
		return 0;
	}

	// Let the default window procedure handle 
	// any message that we don't care about
	return ::DefWindowProc(hWnd, msg, wParam, lParam);
}

